import os
from concurrent.futures import ThreadPoolExecutor
from typing import List

from tqdm import tqdm

try:
    import litellm
except ImportError:
    raise ImportError(
        "litellm is required for LiteLLMBackend. Install with: pip install minieval[litellm]"
    )

from ..datatypes import LMOutput, LMRequest, SamplingParams


class LiteLLMBackend:
    API_KEY_MAPPINGS = {
        "OPENAI_API_KEY": "openai_api_key",
        "ANTHROPIC_API_KEY": "anthropic_api_key",
        "COHERE_API_KEY": "cohere_api_key",
        "REPLICATE_API_TOKEN": "replicate_api_token",
        "HUGGINGFACE_API_KEY": "huggingface_api_key",
        "TOGETHER_API_KEY": "together_api_key",
        "AZURE_API_KEY": "azure_api_key",
        "AZURE_API_BASE": "azure_api_base",
        "AZURE_API_VERSION": "azure_api_version",
        "BEDROCK_AWS_ACCESS_KEY_ID": "aws_access_key_id",
        "BEDROCK_AWS_SECRET_ACCESS_KEY": "aws_secret_access_key",
        "BEDROCK_AWS_REGION_NAME": "aws_region_name",
    }

    def __init__(self, model_name: str, **kwargs):
        self.model_name = model_name
        self.kwargs = kwargs
        self._setup_api_keys()

    def _setup_api_keys(self):
        for env_var, litellm_key in self.API_KEY_MAPPINGS.items():
            value = os.getenv(env_var)
            if value:
                setattr(litellm, litellm_key, value)

    def generate(
        self, requests: List[LMRequest], sampling_params: SamplingParams = None
    ) -> List[List[LMOutput]]:
        # Extract parameters from SamplingParams
        n = sampling_params.repeats if sampling_params.repeats is not None else 1
        max_completion_tokens = sampling_params.max_gen_toks
        temperature = sampling_params.temperature
        logprobs = sampling_params.logprobs if sampling_params.logprobs is not None else True

        # OpenAI API only allows up to 4 stop sequences
        stop_sequences = None
        if sampling_params.stop_sequences:
            if len(sampling_params.stop_sequences) > 4:
                import warnings
                warnings.warn("LiteLLM/OpenAI API only supports up to 4 stop sequences. Truncating stop_sequences to first 4.")
            stop_sequences = sampling_params.stop_sequences[:4]

        def _generate_single(request):
            # Use request context directly as messages (it's already a list of message dicts)
            messages = request.context
            
            response = litellm.completion(
                model=self.model_name,
                messages=messages,
                n=n,
                max_completion_tokens=max_completion_tokens,
                temperature=temperature,
                stop=stop_sequences,
                logprobs=logprobs,
                **self.kwargs,
            )

            results = []
            for choice in response.choices:
                text = choice.message.content
                logprobs_data = getattr(choice, "logprobs", None)
                
                # Convert logprobs to serializable format
                serializable_logprobs = None
                if logprobs_data and hasattr(logprobs_data, 'content'):
                    serializable_logprobs = []
                    for token_logprob in logprobs_data.content:
                        token_data = {
                            'token': token_logprob.token,
                            'logprob': token_logprob.logprob,
                            'bytes': token_logprob.bytes,
                            'top_logprobs': token_logprob.top_logprobs
                        }
                        serializable_logprobs.append(token_data)
                
                results.append(LMOutput(text=text, logprobs=serializable_logprobs))
            return results

        with ThreadPoolExecutor() as executor:
            results = list(
                tqdm(
                    executor.map(_generate_single, requests),
                    total=len(requests),
                    desc="Generating responses",
                )
            )

        return results

    def logprobs(
        self,
        requests: List[LMRequest],
    ) -> List[List[LMOutput]]:
        def _logprobs_single(request):
            if isinstance(request.context, list) and len(request.context) > 0:
                context_str = request.context[0].get("content", "")
            elif isinstance(request.context, str):
                context_str = request.context
            else:
                context_str = ""
                
            response = litellm.completion(
                model=self.model_name,
                messages=[{"role": "user", "content": context_str}],
                max_completion_tokens=50,  # Get enough tokens to see the choice
                temperature=0.0,
                logprobs=True,
                **self.kwargs,
            )
            
            completion_logprobs = []
            
            if response.choices and len(response.choices) > 0:
                choice = response.choices[0]
                
                if hasattr(choice, 'logprobs') and choice.logprobs:
                    logprobs_data = choice.logprobs
                    if hasattr(logprobs_data, 'content') and logprobs_data.content:
                        for token_logprob in logprobs_data.content:
                            if hasattr(token_logprob, 'logprob'):
                                completion_logprobs.append({
                                    'token': getattr(token_logprob, 'token', ''),
                                    'logprob': token_logprob.logprob,
                                    'bytes': getattr(token_logprob, 'bytes', None),
                                    'top_logprobs': getattr(token_logprob, 'top_logprobs', [])
                                })
            
            results = []
            for continuation in request.continuation:
                best_logprob = max(lp['logprob'] for lp in completion_logprobs[:5])  # Use top 5 tokens
                
                # Create logprob entry
                logprob_entry = {
                    'token': continuation[:10] if continuation else '',
                    'logprob': best_logprob,
                    'bytes': None,
                    'top_logprobs': []
                }
                
                results.append(LMOutput(text=continuation, logprobs=[logprob_entry]))
            
            return results

        with ThreadPoolExecutor() as executor:
            results = list(
                tqdm(
                    executor.map(_logprobs_single, requests),
                    total=len(requests),
                    desc="Generating logprobs",
                )
            )

        return results
