from dataclasses import dataclass
from typing import Optional
import jinja2

from minieval.datatypes import Formatter, Instance, LMRequest, RequestType


def base_model_chat_template(messages, tokenize=False):
    assert not tokenize
    assert (
        len(messages) == 1
    ), "Basic chat template currently only supports single-turn conversations!"

    message = messages[0]["content"]
    prompt = f"Question: {message}\nAnswer:"
    return prompt


@dataclass
class CoT(Formatter):
    REQUEST_TYPE: RequestType = RequestType.GENERATE
    template: str = """
{%- if few_shot_examples -%}
{%- for example in few_shot_examples -%}
Question: {{ example.question }}
{% if example.choices -%}
Choices:
{% for choice in example.choices -%}
{{ loop.index0 | int | chr_add(65) }}. {{ choice }}
{% endfor -%}
{% endif -%}
Answer: {{ example.gold_completion }}

{% endfor -%}
{%- endif -%}
{%- if instruction -%}
{{ instruction }}
{%- endif -%}
Question: {{ question }}
{%- if choices %}
Choices:
{{ choices_text }}
{%- endif %}
Answer:""".strip()
    instruction: Optional[str] = None

    def _build_message(self, instance: Instance) -> LMRequest:
        env = jinja2.Environment()
        env.filters['chr_add'] = lambda x, y: chr(x + y)
        template = env.from_string(self.template)

        choices_text = None
        if instance.choices:
            choices_text = "\n".join(
                f"{chr(65+i)}. {choice}" 
                for i, choice in enumerate(instance.choices)
            )

        content = template.render(
            question = instance.question,
            choices = instance.choices,
            few_shot_examples = (self.few_shot or []),
            instruction = self.instruction,
            choices_text = choices_text,
        )

        return LMRequest(messages=[{"role": "user", "content": content}])


@dataclass
class MC(Formatter):
    REQUEST_TYPE: RequestType = RequestType.LOGPROBS
    template: str = """
{%- if few_shot_examples -%}
{%- for example in few_shot_examples -%}
{{ example.question }}
Choices:
{% for choice in example.choices -%}
{{ loop.index0 | int | chr_add(65) }}. {{ choice }}
{% endfor -%}
Answer: {{ (example.solution | int | chr_add(65)) if example.solution is not none else example.gold_completion }}

{% endfor -%}
{%- endif -%}
{{ question }}
Choices:
{{ choices_text }}""".strip()

    def _build_message(self, instance: Instance) -> LMRequest:
        env = jinja2.Environment()
        env.filters['chr_add'] = lambda x, y: chr(x + y)
        template = env.from_string(self.template)

        choices_text = None
        if instance.choices:
            choices_text = "\n".join(f"{chr(65+i)}. {c}" for i, c in enumerate(instance.choices))
            
        content = template.render(
            question=instance.question,
            choices=instance.choices,
            choices_text=choices_text,
            few_shot_examples=self.few_shot or [],
            instance=instance
        )
        
        # Build request with continuations " A", " B", etc.
        choices = [f" {chr(65+i)}" for i in range(len(instance.choices))] if instance.choices else None

        return LMRequest(
            messages=[{"role": "user", "content": content}],
            continuation=choices
        )


@dataclass
class RC(Formatter):
    REQUEST_TYPE: RequestType = RequestType.LOGPROBS
    template: str = """
{%- if few_shot_examples -%}
{%- for example in few_shot_examples -%}
{{ example.question }}
Choices:
{% for choice in example.choices -%}
{{ loop.index0 | int | chr_add(65) }}. {{ choice }}
{% endfor -%}
Answer: {{ example.gold_completion }}

{% endfor -%}
{%- endif -%}
{{ question }}
Choices:
{{ choices_text }}""".strip()

    def _build_message(self, instance: Instance) -> LMRequest:
        env = jinja2.Environment()
        env.filters['chr_add'] = lambda x, y: chr(x + y)
        template = env.from_string(self.template)

        # Add choices
        choices_text = None
        choice_letters = None
        if instance.choices:
            choices_text = "\n".join(
                f"{chr(65+i)}. {choice}" for i, choice in enumerate(instance.choices)
            )
            choice_letters = [chr(65+i) for i in range(len(instance.choices))]
        
        content = template.render(
            question = instance.question,
            choices = instance.choices,
            few_shot_examples = (self.few_shot or []),
            instance = instance,
            choices_text = choices_text,
            choice_letters = choice_letters
        )
        
        messages = [{"role": "user", "content": content}]
        
        # Add leading space to continuations
        choices = [" " + choice for choice in instance.choices]

        request = LMRequest(messages=messages, continuation=choices)
        return request


@dataclass
class Continuation(Formatter):
    REQUEST_TYPE: RequestType = RequestType.LOGPROBS
    template: str = """
{%- if few_shot_examples -%}
{%- for example in few_shot_examples -%}
{{ example.question }}
Answer: {{ example.gold_completion }}

{% endfor -%}
{%- endif -%}
{{ question }}""".strip()

    def _build_message(self, instance: Instance) -> LMRequest:
        template = jinja2.Template(self.template)

        content = template.render(
            question=instance.question,
            choices=instance.choices, 
            few_shot_examples=self.few_shot or [],
            instance=instance
        )

        # Add leading spaces for continuations
        choices = [f" {choice}" for choice in (instance.choices or [])]
        
        return LMRequest(
            messages=[{"role": "user", "content": content}],
            continuation=choices
        )


@dataclass
class PPL(Formatter):
    REQUEST_TYPE: RequestType = RequestType.LOGPROBS
    template: Optional[str] = None

    def __post_init__(self):
        assert self.few_shot_n == 0, "PPL formatter does not support few shot examples"

    def _build_message(self, instance: Instance) -> LMRequest:
        messages = [{"role": "user", "content": ""}]
        request = LMRequest(messages=messages, continuation=instance.gold_completion)
        return request
