from dataclasses import dataclass

from minieval.datatypes import Metric, Response


@dataclass
class LogprobGold(Metric):
    """Logprob of the correct continuation"""

    name: str = "gold_logprob"

    def _compute_metric_single(self, response: Response, scorer: list) -> float:
        response.scores[self.name] = {}

        gold_idx = response.input.solution

        assert isinstance(gold_idx, int), "Solution must be the index of the correct completion!"

        scores = [output.score[scorer] for output in response.output]

        score = scores[gold_idx]

        return score


@dataclass
class LogprobAccuracy(Metric):
    """Compute accuracy by selecting from N continuations with the highest logprob (e.g., " A" vs. " B" vs. ...)"""

    name: str = "acc"

    def _compute_metric_single(self, response: Response, scorer: list) -> bool:

        gold_idx = response.input.solution

        assert isinstance(gold_idx, int), "Solution must be the index of the correct completion!"

        scores = [output.score[scorer] for output in response.output]

        max_idx = max(range(len(scores)), key=lambda i: scores[i])

        score = float(max_idx == gold_idx)

        return score


@dataclass
class Top1(Metric):
    """Check whether a single answer matches a solution"""
    name: str = "top1"

    def _compute_metric_single(self, response: Response, scorer: list) -> Response:
        assert len(response.output) == 1, "Accuracy only supports single generations"

        scores = [output.score[scorer] for output in response.output]

        score = scores[0]

        return score


@dataclass
class PassAtK(Metric):
    """ pass@k originally introduced in https://arxiv.org/abs/2107.03374 """
    k: int = 1
    name: str = None

    def __post_init__(self):
        self.name = f"pass@{self.k}"

    def _compute_metric_single(self, response: Response, scorer: list) -> Response:
        assert (
            len(response.output) >= self.k
        ), f"Cannot compute pass@k when n < k. n={len(response.output)}, k={self.k}"

        scores = [output.score[scorer] for output in response.output]

        # @DAVIDH THIS IS NOT CORRECT NEED TO FIX

        pass_at_k = any(scores[: self.k])

        return pass_at_k


@dataclass
class MajAtK(Metric):
    """ A majority vote of N samples """
    k: int = 1
    name: str = None

    def __post_init__(self):
        self.name = f"maj@{self.k}"

    def _compute_metric_single(self, response: Response, scorer: str) -> Response:
        assert (
            len(response.output) >= self.k
        ), f"Cannot compute pass@k when n < k. n={len(response.output)}, k={self.k}"

        # Get all extracted answers and their counts for this scorer
        answer_counts = {}
        for output in response.output[: self.k]:
            ans = str(output.extracted_answer)
            if ans not in answer_counts:
                answer_counts[ans] = 0
            answer_counts[ans] += 1

        # Find most common answer
        majority_answer = max(answer_counts.items(), key=lambda x: x[1])[0]

        # Get score for majority answer
        majority_outputs = [
            output
            for output in response.output[: self.k]
            if str(output.extracted_answer) == majority_answer
        ]

        # All outputs with same answer should have same score
        scores = [output.score[scorer] for output in majority_outputs]
        assert len(set(scores)) == 1, "Scores differ for same extracted answer"

        maj_at_k = scores[0]

        return maj_at_k
