from enum import Enum

from minieval.datatypes import Instance, LMOutput, Scorer
from minieval.extract import math_latex


class ScoreType(Enum):
    CONTINUOUS = "continuous"
    BINARY = "binary"


class Logprob(Scorer):
    name = "logprob"
    _type = ScoreType.CONTINUOUS

    def _score_response_single(self, input: Instance, output: LMOutput) -> LMOutput:
        logprobs = output.logprobs

        assert logprobs is not None

        logprobs = [tok['logprob'] for tok in logprobs]

        total_logprob = sum(logprobs)
        output.score[self.name] = total_logprob
        return output


class LogprobPerChar(Scorer):
    name = "logprob_per_char"
    _type = ScoreType.CONTINUOUS

    def _score_response_single(self, input: Instance, output: LMOutput) -> LMOutput:
        logprobs = output.logprobs
        num_chars = len(output.text)

        assert logprobs is not None

        logprobs = [tok['logprob'] for tok in logprobs]

        logprob_per_char = sum(logprobs) / num_chars
        output.score[self.name] = logprob_per_char
        return output


class BitsPerByte(Scorer):
    name = "bits_per_byte"
    _type = ScoreType.CONTINUOUS

    def _score_response_single(self, input: Instance, output: LMOutput) -> LMOutput:
        logprobs = output.logprobs
        num_bytes = len(output.text.encode("utf-8"))

        assert logprobs is not None

        logprobs = [tok['logprob'] for tok in logprobs]

        logprob_per_char = sum(logprobs) / num_bytes
        output.score[self.name] = logprob_per_char
        return output


class Accuracy(Scorer):
    name = "accuracy"
    _type = ScoreType.BINARY

    def _score_response_single(self, input: Instance, output: LMOutput) -> LMOutput:
        answer = output.extracted_answer
        correct = input.solution

        assert isinstance(answer, int) and isinstance(
            correct, int
        ), f"Accuracy requires an answer index! Seeing: {answer=}, {correct=}"

        output.score[self.name] = float(answer == correct)
        return output


class ExactMatch(Scorer):
    name = "exact_match"
    _type = ScoreType.BINARY

    def _score_response_single(self, input: Instance, output: LMOutput) -> LMOutput:
        answer = output.extracted_answer
        correct = input.solution

        assert not isinstance(answer, list), f"EM requires a list of answers! Seeing: {answer}"

        assert isinstance(answer, str), f"EM requires a string answer! Seeing: {answer}"

        output.score[self.name] = float(answer == correct)
        return output


class ExactMatchFlex(Scorer):
    """Allow any extracted answer to be correct"""

    name = "exact_match_flex"
    _type = ScoreType.BINARY

    def _score_response_single(self, input: Instance, output: LMOutput) -> LMOutput:
        gen_answers = output.extracted_answer
        correct = input.solution

        assert isinstance(
            gen_answers, list
        ), f"EM Flex requires a list of answers! Seeing: {gen_answers}"

        assert all(
            isinstance(gen, str) for gen in gen_answers
        ), f"EM Flex requires a list of strings! Seeing: {gen_answers}"

        for gen in gen_answers:
            if math_latex.is_equiv(gen, correct):
                output.score[self.name] = 1.0
                return output

        output.score[self.name] = 0.0
        return output
