import logging
from typing import Optional

from math_verify import parse, verify, LatexExtractionConfig, ExprExtractionConfig

from minieval.datatypes import Instance, LMOutput, Scorer
from minieval.score.core import ScoreType

logger = logging.getLogger(__name__)


class MathVerify(Scorer):
    """ A wrapper for https://github.com/huggingface/Math-Verify """

    name = "math_verify"
    _type = ScoreType.BINARY

    def __init__(self, 
                 use_latex: bool = True, 
                 use_expr: bool = True,
                 timeout_seconds: int = 10):
        self.use_latex = use_latex
        self.use_expr = use_expr
        self.timeout_seconds = timeout_seconds
        
        # Configure extraction methods
        self.extraction_configs = []
        if use_latex:
            self.extraction_configs.append(LatexExtractionConfig())
        if use_expr:
            self.extraction_configs.append(ExprExtractionConfig())
            
        if not self.extraction_configs:
            raise ValueError("At least one extraction method must be enabled")

    def _parse_answer(self, text: str) -> Optional[object]:
        try:
            return parse(text, extraction_config=self.extraction_configs)
        except Exception as e:
            logger.debug(f"Failed to parse '{text}': {e}")
            return None

    def _score_response_single(self, input: Instance, output: LMOutput) -> LMOutput:
        # Get the generated answer text
        if hasattr(output, 'extracted_answer') and output.extracted_answer:
            # If we already have an extracted answer, use it
            if isinstance(output.extracted_answer, list):
                generated_text = output.extracted_answer[0] if output.extracted_answer else output.text
            else:
                generated_text = str(output.extracted_answer)
        else:
            # Otherwise use the full text
            generated_text = output.text
            
        # Get the gold answer
        gold_text = str(input.solution)
        
        # Parse both answers
        try:
            parsed_gold = self._parse_answer(gold_text)
            parsed_generated = self._parse_answer(generated_text)
            
            # If either parsing failed, score as 0
            if parsed_gold is None or parsed_generated is None:
                logger.debug(f"Parsing failed - gold: {parsed_gold is not None}, generated: {parsed_generated is not None}")
                output.score[self.name] = 0.0
                return output
                
            # Use Math-Verify's verify function to compare
            # Note: order matters - verify(gold, generated) 
            is_correct = verify(parsed_gold, parsed_generated)
            output.score[self.name] = float(is_correct)
            
            logger.debug(f"Math-Verify comparison: '{gold_text}' vs '{generated_text}' -> {is_correct}")
            
        except Exception as e:
            logger.warning(f"Math-Verify verification failed: {e}")
            output.score[self.name] = 0.0
            
        return output