from minieval.datatypes import TaskConfig, TaskRegistry, SamplingParams
from minieval.few_shot import FewShotRegistry
from minieval.formatters import MC, RC, Continuation, CoT
from minieval.metrics import LogprobAccuracy, LogprobGold, PassAtK, Top1
from minieval.score.core import (
    BitsPerByte,
    ExactMatch,
    ExactMatchFlex,
    Logprob,
    LogprobPerChar,
)
from minieval.score.code import CodeExecution
from minieval.score.math_verify import MathVerify
from minieval.tasks.aime import AIME
from minieval.tasks.arc import ARCEasy
from minieval.tasks.gpqa import GPQA
from minieval.tasks.gsm import GSM8K
from minieval.tasks.hellaswag import HellaSwag
from minieval.tasks.humaneval import HumanEval
from minieval.tasks.minerva import Math500, MinervaMath
from minieval.tasks.mmlu import MMLU, MMLUPro


@TaskRegistry.register("arc_easy:mc", task=ARCEasy)
class ARCEasyMC(TaskConfig):
    formatter = MC(few_shot_alias=FewShotRegistry.get("arc_easy::olmes"), few_shot_n=5)
    scorer = [Logprob()]
    metric = [LogprobAccuracy(), LogprobGold()]


@TaskRegistry.register("arc_easy:rc", task=ARCEasy)
class ARCEasyRC(TaskConfig):
    formatter = RC(few_shot_alias=FewShotRegistry.get("arc_easy::olmes"), few_shot_n=5)
    # extractor=[CoreExtract.strip_think_toks, ARCEasy.extract_answer]
    scorer = [LogprobPerChar(), BitsPerByte()]
    metric = [LogprobAccuracy(), LogprobGold()]


@TaskRegistry.register("hellaswag", task=HellaSwag)
class HellaSwag(TaskConfig):
    formatter = Continuation(few_shot_alias=FewShotRegistry.get("hellaswag::olmes"), few_shot_n=5)
    scorer = [LogprobPerChar(), BitsPerByte()]
    metric = [LogprobAccuracy(), LogprobGold()]
    limit = 10_000


for subset in MinervaMath.subsets:

    @TaskRegistry.register(f"minerva_{subset}:cot", task=MinervaMath)
    class MinervaCoT(TaskConfig):
        subset = subset
        formatter = CoT(instruction="Present the answer in LaTex format: \\boxed{Your answer}")
        scorer = [ExactMatchFlex()]
        metric = [Top1()]
        sampling_params = SamplingParams(
            max_gen_toks=2048,
            temperature=0.0,
        )

    @TaskRegistry.register(f"minerva_{subset}:selfc", task=MinervaMath)
    class MinervaSelfC(TaskConfig):
        subset = subset
        formatter = CoT(instruction="Present the answer in LaTex format: \\boxed{Your answer}")
        scorer = [ExactMatchFlex()]
        metric = [PassAtK(k=1), PassAtK(k=2), PassAtK(k=5), PassAtK(k=20)]
        sampling_params = SamplingParams(max_gen_toks=2048, temperature=0.0, repeats=20)


@TaskRegistry.register("minerva_500:few_shot_cot", task=Math500)
class Minerva500CoT(TaskConfig):
    formatter = CoT(
        few_shot_alias=FewShotRegistry.get("minerva::oe_eval"),
        few_shot_n=5,
    )
    scorer = [ExactMatchFlex()]
    metric = [Top1()]
    sampling_params = SamplingParams(
        max_gen_toks=2048,
        temperature=0.0,
    )


@TaskRegistry.register("minerva_500:cot", task=Math500)
class Minerva500CoT(TaskConfig):
    formatter = CoT(instruction="Present the answer in LaTex format: \\boxed{Your answer}")
    scorer = [ExactMatchFlex()]
    metric = [Top1()]
    sampling_params = SamplingParams(
        max_gen_toks=2048,
        temperature=0.0,
    )


@TaskRegistry.register("minerva_500:math_verify", task=Math500)
class Minerva500MathVerify(TaskConfig):
    formatter = CoT(instruction="Present the answer in LaTex format: \\boxed{Your answer}")
    scorer = [MathVerify()]
    metric = [Top1()]
    sampling_params = SamplingParams(
        max_gen_toks=2048,
        temperature=0.0,
    )


@TaskRegistry.register("gpqa:cot", task=GPQA)
class GPQACoT(TaskConfig):
    formatter = CoT(
        instruction='Answer the following multiple-choice question by giving the correct answer letter in parentheses. Provide CONCISE reasoning for the answer, and make sure to finish the response with "Therefore, the answer is (ANSWER_LETTER)" where (ANSWER_LETTER) is one of (A), (B), (C), (D).\n\n'
    )
    scorer = [ExactMatch()]
    metric = [Top1()]
    sampling_params = SamplingParams(
        max_gen_toks=8192,
        temperature=0.0,
    )


@TaskRegistry.register("gsm:cot", task=GSM8K)
class GSMCoT(TaskConfig):
    formatter = CoT()
    scorer = [ExactMatchFlex()]
    metric = [Top1()]
    sampling_params = SamplingParams(
        max_gen_toks=2048,
        temperature=0.0,
    )


@TaskRegistry.register("humaneval", task=HumanEval)
class HumanEval(TaskConfig):
    formatter = CoT(
        instruction="Here is the completed function:\n\n```python\n"
    )
    scorer = [CodeExecution(timeout=20)]
    metric = [PassAtK(k=16)]
    sampling_params = SamplingParams(
        max_gen_toks=1024,
        temperature=0.8,
        top_p=0.95,
        repeats=16,
        stop_sequences=[
            "\nclass", "\nif", "\nprint", "\n#", "\n```", "\n```\n\n", "<|eot_id|>",
        ]
    )


@TaskRegistry.register("humaneval:bpb", task=HumanEval)
class HumanEvalBPB(TaskConfig):
    formatter = CoT()
    scorer = [BitsPerByte()]
    metric = [LogprobGold()]


@TaskRegistry.register("aime:cot", task=AIME)
class AIMECoT(TaskConfig):
    formatter = CoT(instruction="Present the answer in LaTex format: \\boxed{Your answer}")
    scorer = [ExactMatchFlex()]
    metric = [Top1()]
    sampling_params = SamplingParams(
        max_gen_toks=8192,
        temperature=0.0,
    )


for year in AIME.subsets:

    @TaskRegistry.register(f"aime_{year}:cot", task=AIME)
    class AIMECoT(TaskConfig):
        subset = year
        formatter = CoT(instruction="Present the answer in LaTex format: \\boxed{Your answer}")
        scorer = [ExactMatchFlex()]
        metric = [Top1()]
        sampling_params = SamplingParams(
            max_gen_toks=8192,
            temperature=0.0,
        )


for subset in MMLU.subsets:

    @TaskRegistry.register(f"mmlu_{subset}:mc", task=MMLU)
    class MMLUMC(TaskConfig):
        subset = subset
        formatter = MC(
            few_shot_n=5,
        )
        scorer = [Logprob()]
        metric = [LogprobAccuracy(), LogprobGold()]
        sampling_params = SamplingParams(
            max_gen_toks=2048,
            temperature=0.0,
            stop_sequences=["Question:"],
        )

    @TaskRegistry.register(f"mmlu_{subset}:cot", task=MMLU)
    class MMLUCoT(TaskConfig):
        subset = subset
        formatter = CoT(
            instruction='Answer the following multiple-choice question by giving the correct answer letter in parentheses. Provide CONCISE reasoning for the answer, and make sure to finish the response with "Therefore, the answer is (ANSWER_LETTER)" where (ANSWER_LETTER) is one of (A), (B), (C), (D).\n\n'
        )
        scorer = [ExactMatch()]
        metric = [Top1()]
        sampling_params = SamplingParams(
            max_gen_toks=2048,
            temperature=0.0,
            stop_sequences=["Question:"],
        )


for subset in MMLUPro.subsets:

    @TaskRegistry.register(f"mmlu_pro_{subset}:cot", task=MMLUPro)
    class MMLUProCoT(TaskConfig):
        subset = subset
        formatter = CoT(
            instruction='Answer the following multiple-choice question by giving the correct answer letter in parentheses. Provide CONCISE reasoning for the answer, and make sure to finish the response with "Therefore, the answer is (ANSWER_LETTER)" where (ANSWER_LETTER) is one of (A), (B), (C), (D).\n\n'
        )
        scorer = [ExactMatch()]
        metric = [Top1()]
        sampling_params = SamplingParams(
            max_gen_toks=2048,
            temperature=0.0,
            stop_sequences=["Question:"],
        )
