import random
import re

from datasets import load_dataset

from minieval.datatypes import Instance, LMOutput, Response, Task, TaskConfig
from minieval.extract import qa


class GPQA(Task):
    hf_path = "Idavidrein/gpqa"

    def __init__(self, config: TaskConfig):
        self.config = config
        self._requests = None

    @property
    def requests(self):
        if self._requests is None:
            dataset = load_dataset(path=self.hf_path, name="gpqa_main", split="train")
            self._requests = list(map(self._process_instance, dataset))
        return self._requests

    def _process_instance(self, doc):
        gold_answer = self._preprocess(doc["Correct Answer"])
        choices = [
            self._preprocess(doc["Incorrect Answer 1"]),
            self._preprocess(doc["Incorrect Answer 2"]),
            self._preprocess(doc["Incorrect Answer 3"]),
            gold_answer,
        ]

        random.Random(self.config.seed + hash(doc["Record ID"])).shuffle(choices)
        correct_answer_index = choices.index(gold_answer)

        return Instance(
            question=doc["Question"],
            gold_completion=choices[correct_answer_index],
            choices=choices,
            solution=["A", "B", "C", "D"][correct_answer_index],
            metadata={
                "id": doc["Record ID"],
                "canary_string": doc["Canary String"],
                "explanation": doc["Explanation"],
            },
        )

    @classmethod
    def extract_answer(self, response: Response, generation: LMOutput) -> str:
        answer = qa.extract_mcqa_answer(
            generation.text, answer_regexes=["\\(?([A-D])\\)?"]  # both "(A)" and "A"
        )
        if answer in ["A", "B", "C", "D"]:
            return answer
        return None

    def _preprocess(self, text):
        if text is None:
            return " "
        text = text.strip()
        text = text.replace(" [title]", ". ")
        text = re.sub("\\[.*?\\]", "", text)
        text = text.replace("  ", " ")
        return text


class GPQADiamond(GPQA):
    def __init__(self, config: TaskConfig):
        self.config = config
        self._requests = None

    @property
    def requests(self):
        if self._requests is None:
            dataset = load_dataset(path=self.hf_path, name="gpqa_diamond", split="train")
            self._requests = list(map(self._process_instance, dataset))
        return self._requests


class SuperGPQA(GPQA):
    hf_path = "m-a-p/SuperGPQA"

    def __init__(self, config: TaskConfig):
        self.config = config
        self._requests = None

    @property
    def requests(self):
        if self._requests is None:
            dataset = load_dataset(path=self.hf_path, split="train")
            self._requests = list(map(self._process_instance, dataset))
        return self._requests

    def _process_instance(self, doc):
        gold_answer = doc["answer"]
        choices: list[str] = doc["options"]

        correct_answer_index = choices.index(gold_answer)

        return Instance(
            question=doc["question"],
            gold_completion=gold_answer,
            choices=doc["options"],
            solution=["A", "B", "C", "D"][correct_answer_index],
            metadata={
                "id": doc["uuid"],
                "discipline": doc["discipline"],
                "field": doc["field"],
                "subfield": doc["subfield"],
                "difficulty": doc["difficulty"],
                "is_calculation": doc["is_calculation"],
            },
        )
