import logging
import os
import sys
import json
from typing import List, Dict, Any
import requests
from rich.console import Console
from rich.table import Table
from rich.json import JSON

logger = logging.getLogger(__name__)
console = Console()


class DatalakeAPI:
    """Basic utilities for interacting with the datalake API"""
    
    def __init__(self, base_url: str = "https://oe-eval-datalake.allen.ai"):
        self.base_url = base_url
    
    def inspect_experiment(self, experiment_id: str) -> Dict[str, Any]:
        """Check if experiment files exist in greenlake/inspect/"""
        url = f"{self.base_url}/greenlake/inspect/{experiment_id}"
        try:
            response = requests.get(url)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed to inspect experiment {experiment_id}: {e}")
            raise
    
    def get_metrics_all(self, experiment_id: str) -> Dict[str, Any]:
        """Pull metrics from greenlake/metrics-all/"""
        url = f"{self.base_url}/greenlake/metrics-all/{experiment_id}"
        try:
            response = requests.get(url)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed to get metrics for experiment {experiment_id}: {e}")
            raise
    
    def find_experiments(self, **kwargs) -> List[Dict[str, Any]]:
        """Find experiments using bluelake/find-experiments/"""
        url = f"{self.base_url}/bluelake/find-experiments/"
        try:
            response = requests.get(url, params=kwargs)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed to find experiments: {e}")
            raise
    
    def download_file(self, experiment_id: str, task_idx: int, result_type: str) -> bytes:
        """Download a specific file from the datalake"""
        url = f"{self.base_url}/greenlake/download-result/{experiment_id}"
        params = {"task_idx": task_idx, "resulttype": result_type}
        
        try:
            headers = {
                'Accept-Encoding': 'gzip, deflate',
                'User-Agent': 'minieval-datalake-client/1.0'
            }
            
            response = requests.get(url, params=params, timeout=30, headers=headers)
            response.raise_for_status()
            return response.content
        except requests.exceptions.RequestException as e:
            logger.error(f"Failed to download {result_type} for experiment {experiment_id}: {e}")
            raise


class DatalakeDownloader:
    """Simple downloader to pull all result files to a specified output folder"""
    
    def __init__(self, api: DatalakeAPI):
        self.api = api
    
    def download_all_files(self, experiment_id: str, output_dir: str, task_idx: int = 0) -> List[str]:
        """Download all available files for an experiment"""
        # First inspect to see what files are available
        try:
            inspection = self.api.inspect_experiment(experiment_id)
            console.print(f"[bold blue]Experiment {experiment_id} file counts:[/bold blue]")
            
            table = Table(title="Available Files")
            table.add_column("File Type", style="cyan")
            table.add_column("Count", style="magenta")
            
            available_types = []
            for file_type in ["ALL_METRICS", "PREDICTIONS", "INPUTS", "METADATA"]:
                count = inspection.get(f"{file_type}_files", 0)
                table.add_row(file_type, str(count))
                if count > 0:
                    available_types.append(file_type)
            
            console.print(table)
            
            if not available_types:
                console.print("[bold red]No files found in experiment![/bold red]")
                return []
            
        except Exception as e:
            console.print(f"[bold yellow]Warning: Could not inspect experiment: {e}[/bold yellow]")
            # Try to download common file types anyway
            available_types = ["PREDICTIONS", "INPUTS"]
        
        # Create output directory
        os.makedirs(output_dir, exist_ok=True)
        
        downloaded_files = []
        
        # Download each available file type
        for result_type in available_types:
            try:
                content = self.api.download_file(experiment_id, task_idx, result_type)
                
                if content:
                    filename = f"task-{task_idx:03d}-{result_type.lower()}.jsonl"
                    if result_type == "METRICS":
                        raise NotImplementedError("METRICS not supported in datalake API.")
                    
                    filepath = os.path.join(output_dir, filename)
                    with open(filepath, 'wb') as f:
                        f.write(content)
                    
                    downloaded_files.append(filename)
                    console.print(f"[green]✓[/green] Downloaded {filename}")
                else:
                    console.print(f"[yellow]![/yellow] Empty content for {result_type}")
                    
            except Exception as e:
                console.print(f"[red]✗[/red] Failed to download {result_type}: {e}")
        
        return downloaded_files


class DatalakeInspector:
    """Simple command to inspect the first output from a job"""
    
    def __init__(self, api: DatalakeAPI):
        self.api = api
    
    def inspect_experiment_output(self, experiment_id: str, task_idx: int = 0, show_first_n: int = 3):
        """Inspect the first few outputs from an experiment"""
        console.print(f"[bold blue]Inspecting experiment {experiment_id}[/bold blue]")
        
        # Get basic experiment info
        try:
            inspection = self.api.inspect_experiment(experiment_id)
            console.print(f"[bold]File counts:[/bold]")
            for key, value in inspection.items():
                if key.endswith("_files"):
                    console.print(f"  {key}: {value}")
        except Exception as e:
            console.print(f"[yellow]Could not get file counts: {e}[/yellow]")
        
        # Try to get and display metrics
        try:
            metrics = self.api.get_metrics_all(experiment_id)
            console.print(f"\n[bold]Metrics:[/bold]")
            console.print(JSON.from_data(metrics))
        except Exception as e:
            console.print(f"[yellow]Could not get metrics: {e}[/yellow]")
        
        # Try to download and show first few predictions
        try:
            predictions_content = self.api.download_file(experiment_id, task_idx, "PREDICTIONS")
            if predictions_content:
                console.print(f"\n[bold]First {show_first_n} predictions:[/bold]")
                
                # Parse JSONL content
                lines = predictions_content.decode('utf-8').strip().split('\n')
                for i, line in enumerate(lines[:show_first_n]):
                    if line.strip():
                        try:
                            prediction = json.loads(line)
                            console.print(f"\n[bold cyan]Prediction {i+1}:[/bold cyan]")
                            console.print(JSON.from_data(prediction))
                        except json.JSONDecodeError:
                            console.print(f"[red]Could not parse prediction {i+1}[/red]")
            else:
                console.print("[yellow]No predictions found[/yellow]")
                
        except Exception as e:
            console.print(f"[yellow]Could not get predictions: {e}[/yellow]")


def main():
    import argparse

    # Require invocation "minieval datalake ..."
    import sys
    if not (len(sys.argv) >= 2 and sys.argv[1] == "datalake"):
        sys.stderr.write(
            "[red]This CLI must be invoked as: minieval datalake ...\n"
            "Example: minieval datalake --help\n[/red]"
        )
        sys.exit(1)
    sys_argv = sys.argv[:1] + sys.argv[2:]


    parser = argparse.ArgumentParser(description="Datalake utilities for experiment inspection and downloading")
    parser.add_argument("--base-url", default="https://oe-eval-datalake.allen.ai", help="Datalake base URL")

    subparsers = parser.add_subparsers(dest='command', help='Available commands')

    # Inspect command
    inspect_parser = subparsers.add_parser('inspect', help='Inspect experiment files and show first outputs')
    inspect_parser.add_argument('experiment_id', help='Experiment ID to inspect')
    inspect_parser.add_argument('--task-idx', type=int, default=0, help='Task index (default: 0)')
    inspect_parser.add_argument('--show-first-n', type=int, default=3, help='Number of first outputs to show (default: 3)')

    # Download command
    download_parser = subparsers.add_parser('download', help='Download all files for an experiment')
    download_parser.add_argument('experiment_id', help='Experiment ID to download')
    download_parser.add_argument('output_dir', nargs='?', default='/tmp/datalake_pull', help='Output directory for downloaded files (default: /tmp/datalake_pull)')
    download_parser.add_argument('--task-idx', type=int, default=0, help='Task index (default: 0)')

    # Metrics command
    metrics_parser = subparsers.add_parser('metrics', help='Get metrics for an experiment')
    metrics_parser.add_argument('experiment_id', help='Experiment ID to get metrics for')

    args = parser.parse_args(sys_argv[1:])

    if not args.command:
        parser.print_help()
        return

    # Initialize API
    api = DatalakeAPI(args.base_url)

    try:
        if args.command == 'inspect':
            inspector = DatalakeInspector(api)
            inspector.inspect_experiment_output(args.experiment_id, args.task_idx, args.show_first_n)

        elif args.command == 'download':
            downloader = DatalakeDownloader(api)
            files = downloader.download_all_files(args.experiment_id, args.output_dir, args.task_idx)
            if files:
                console.print(f"\n[bold green]Successfully downloaded {len(files)} files to {args.output_dir}[/bold green]")
            else:
                console.print(f"[bold red]No files downloaded[/bold red]")

        elif args.command == 'metrics':
            metrics = api.get_metrics_all(args.experiment_id)
            console.print(JSON.from_data(metrics))

        elif args.command == 'find':
            raise NotImplementedError("@david -- doesn't work yet. ideally, we can have a command where we specify a model name, workspace name, etc. and can pull results")

    except Exception as e:
        console.print(f"[bold red]Error: {e}[/bold red]")
        sys.exit(1)


if __name__ == '__main__':
    main()